use crate::{
    mesh::{
        Face,
        Side,
    },
    voxel::OPAQUE,
    Chunk,
    Voxel,
};

pub(crate) fn ao_value(side1: bool, corner: bool, side2: bool) -> u32 {
    match (side1, corner, side2) {
        (true, _, true) => 0,
        (true, true, false) | (false, true, true) => 1,
        (false, false, false) => 3,
        _ => 2,
    }
}

pub(crate) fn side_aos<V: Voxel>(neighbors: [V; 8]) -> [u32; 4] {
    let ns = [
        neighbors[0].visibility() == OPAQUE,
        neighbors[1].visibility() == OPAQUE,
        neighbors[2].visibility() == OPAQUE,
        neighbors[3].visibility() == OPAQUE,
        neighbors[4].visibility() == OPAQUE,
        neighbors[5].visibility() == OPAQUE,
        neighbors[6].visibility() == OPAQUE,
        neighbors[7].visibility() == OPAQUE,
    ];

    [
        ao_value(ns[0], ns[1], ns[2]),
        ao_value(ns[2], ns[3], ns[4]),
        ao_value(ns[6], ns[7], ns[0]),
        ao_value(ns[4], ns[5], ns[6]),
    ]
}

pub(crate) fn face_aos<C, V>(face: &Face, chunk: &C) -> [u32; 4]
where
    C: Chunk<Output = V>,
    V: Voxel,
{
    let [x, y, z] = face.voxel();

    match face.side() {
        Side::X_NEG => side_aos([
            chunk.get(x - 1, y, z + 1),
            chunk.get(x - 1, y - 1, z + 1),
            chunk.get(x - 1, y - 1, z),
            chunk.get(x - 1, y - 1, z - 1),
            chunk.get(x - 1, y, z - 1),
            chunk.get(x - 1, y + 1, z - 1),
            chunk.get(x - 1, y + 1, z),
            chunk.get(x - 1, y + 1, z + 1),
        ]),
        Side::X_POS => side_aos([
            chunk.get(x + 1, y, z - 1),
            chunk.get(x + 1, y - 1, z - 1),
            chunk.get(x + 1, y - 1, z),
            chunk.get(x + 1, y - 1, z + 1),
            chunk.get(x + 1, y, z + 1),
            chunk.get(x + 1, y + 1, z + 1),
            chunk.get(x + 1, y + 1, z),
            chunk.get(x + 1, y + 1, z - 1),
        ]),
        Side::Y_NEG => side_aos([
            chunk.get(x - 1, y - 1, z),
            chunk.get(x - 1, y - 1, z + 1),
            chunk.get(x, y - 1, z + 1),
            chunk.get(x + 1, y - 1, z + 1),
            chunk.get(x + 1, y - 1, z),
            chunk.get(x + 1, y - 1, z - 1),
            chunk.get(x, y - 1, z - 1),
            chunk.get(x - 1, y - 1, z - 1),
        ]),
        Side::Y_POS => side_aos([
            chunk.get(x, y + 1, z + 1),
            chunk.get(x - 1, y + 1, z + 1),
            chunk.get(x - 1, y + 1, z),
            chunk.get(x - 1, y + 1, z - 1),
            chunk.get(x, y + 1, z - 1),
            chunk.get(x + 1, y + 1, z - 1),
            chunk.get(x + 1, y + 1, z),
            chunk.get(x + 1, y + 1, z + 1),
        ]),
        Side::Z_NEG => side_aos([
            chunk.get(x - 1, y, z - 1),
            chunk.get(x - 1, y - 1, z - 1),
            chunk.get(x, y - 1, z - 1),
            chunk.get(x + 1, y - 1, z - 1),
            chunk.get(x + 1, y, z - 1),
            chunk.get(x + 1, y + 1, z - 1),
            chunk.get(x, y + 1, z - 1),
            chunk.get(x - 1, y + 1, z - 1),
        ]),
        Side::Z_POS => side_aos([
            chunk.get(x + 1, y, z + 1),
            chunk.get(x + 1, y - 1, z + 1),
            chunk.get(x, y - 1, z + 1),
            chunk.get(x - 1, y - 1, z + 1),
            chunk.get(x - 1, y, z + 1),
            chunk.get(x - 1, y + 1, z + 1),
            chunk.get(x, y + 1, z + 1),
            chunk.get(x + 1, y + 1, z + 1),
        ]),
    }
}
