pub trait Chunk {
    type Output;

    const X: usize;
    const Y: usize;
    const Z: usize;

    #[inline]
    fn size() -> usize {
        Self::X * Self::Y * Self::Z
    }

    #[inline]
    fn linearize(x: usize, y: usize, z: usize) -> usize {
        x + (y * Self::X) + (z * Self::X * Self::Y)
    }

    #[inline]
    fn delinearize(mut index: usize) -> (usize, usize, usize) {
        let z = index / (Self::X * Self::Y);
        index -= z * (Self::X * Self::Y);

        let y = index / Self::X;
        index -= y * Self::X;

        let x = index;

        (x, y, z)
    }

    fn get(&self, x: usize, y: usize, z: usize) -> Self::Output;
}
