use crate::{
    chunk::Chunk,
    mesh::QuadGroups,
    voxel::{
        Voxel,
        EMPTY,
        OPAQUE,
        TRANSPARENT,
    },
};

#[inline]
pub fn simple_mesh<C, T>(chunk: &C) -> QuadGroups
where
    C: Chunk<Output = T>,
    T: Voxel,
{
    let mut buffer = QuadGroups::default();

    simple_mesh_buffer(chunk, &mut buffer);

    buffer
}

#[inline]
pub fn simple_mesh_buffer<C, V>(chunk: &C, buffer: &mut QuadGroups)
where
    C: Chunk<Output = V>,
    V: Voxel,
{
    assert!(C::X >= 2);
    assert!(C::Y >= 2);
    assert!(C::Z >= 2);

    buffer.clear();

    for z in 1..C::Z - 1 {
        for y in 1..C::Y - 1 {
            for x in 1..C::X - 1 {
                let voxel = chunk.get(x, y, z);

                match voxel.visibility() {
                    EMPTY => continue,
                    visibility => {
                        let neighbors = [
                            chunk.get(x - 1, y, z),
                            chunk.get(x + 1, y, z),
                            chunk.get(x, y - 1, z),
                            chunk.get(x, y + 1, z),
                            chunk.get(x, y, z - 1),
                            chunk.get(x, y, z + 1),
                        ];

                        for (i, neighbor) in neighbors.into_iter().enumerate() {
                            let generate = match (visibility, neighbor.visibility()) {
                                (OPAQUE, EMPTY) | (OPAQUE, TRANSPARENT) | (TRANSPARENT, EMPTY) => true,
                                (TRANSPARENT, TRANSPARENT) => voxel != neighbor,
                                (_, _) => false,
                            };

                            if generate {
                                V::mesh(buffer, i, x, y, z);
                            }
                        }
                    }
                }
            }
        }
    }
}
